# Copyright 2016 Huawei Technologies India Pvt. Ltd.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

AGENT_TYPE_BGP_ROUTING = 'BGP dynamic routing agent'

BGP_DRAGENT = 'bgp_dragent'

BGP_PLUGIN = 'q-bgp-plugin'

# List of supported authentication types.
SUPPORTED_AUTH_TYPES = ['none', 'md5']

# Supported AS number range
MIN_ASNUM = 1
MAX_ASNUM = 65535
MAX_4BYTE_ASNUM = 4294967295
